
module ServicesTable exposing (main)

import List
import String exposing (fromInt)
import Browser
import Html exposing (Html, div, span, text, a, table, thead, tbody, tr, th, td)
import Html.Attributes exposing (class, href)
import Http
import Dict exposing (Dict)
import Json.Decode exposing (Decoder, map, map3, field, list, string, dict, keyValuePairs)

pinnedCommit = "2a5f9fcde65629554f39d683db02a98a6c054756"

type alias ServiceInfo =
  { fields : List String
  , tests : List String
  , mod : String }

type alias Info = List
  { name : String
  , fields : List String
  , tests : List String
  , mod : String }

type Model
 = Failure
 | Loading
 | Success Info

type Msg = GotInfo (Result Http.Error Info)

main = Browser.element
  { init = init
  , update = update
  , subscriptions = subscriptions
  , view = view }

init : () -> (Model, Cmd Msg)
init _ =
  ( Loading
  , Http.get
   { url = "https://share.recursor.wf/systemd-services-tests.json"
   , expect = Http.expectJson GotInfo (map sorted dataDecoder) } )

sorted : Info -> Info
sorted info = List.sortBy (\{name, fields, tests, mod} -> (List.length fields, name)) info

dataDecoder : Decoder Info
dataDecoder = map (List.map (\(name, (fields, tests, mod)) -> { name = name, fields = fields, tests = tests, mod = mod }))
  <| keyValuePairs (map3 (\x y z -> (x, y, z)) (field "fields" <| list string) (field "tests" <| list string) (field "module" string))

update : Msg -> Model -> (Model, Cmd Msg)
update msg model =
  case msg of
    GotInfo result ->
      case result of
        Ok info -> (Success info, Cmd.none)
        Err _ -> (Failure, Cmd.none)

subscriptions : Model -> Sub Msg
subscriptions model = Sub.none

-- Render

watchedFields =
  [ "PrivateDevices"
  , "PrivateMounts"
  , "PrivateNetwork"
  , "PrivateTmp"
  , "PrivateUsers"
  , "ProtectControlGroups"
  , "ProtectKernelModules"
  , "ProtectKernelTunables"
  , "ProtectKernelLogs"
  , "ProtectClock"
  , "ProtectHostname"
  , "LockPersonality"
  , "MemoryDenyWriteExecute"
  , "NoNewPrivileges"
  , "RestrictRealtime"
  , "RestrictSUIDSGID" ]

view : Model -> Html Msg
view model = case model of
  Failure -> text "Unable to load data."
  Loading -> text "Loading..."
  Success info -> renderTable info

rotated : String -> Html Msg
rotated txt = th [ class "rotate-45" ] [ div [] [ span [] [ text txt ] ] ]

headers : Html Msg
headers = tr []
  <| [ th [] [ text "source" ], th [] [ text "#" ], th [] [ text "module" ] ]
  ++ List.map rotated watchedFields

renderTable : Info -> Html Msg
renderTable info = table [ class "services-table", class "table-header-rotated" ]
  [ thead [ class "info-thead" ] [ headers ]
  , tbody [] (List.indexedMap renderLine info) ]

mkSourceLink : String -> Html Msg
mkSourceLink mod = a [ href <| "https://github.com/NixOS/nixpkgs/blob/" ++ pinnedCommit ++ String.dropLeft 7 mod ] [ text mod ]

renderLine : Int -> { name : String, fields : List String, tests : List String, mod : String } -> Html Msg
renderLine i {name, fields, tests, mod} = tr []
  <| [ th [] [ mkSourceLink mod ], th [] [ text <| fromInt (i + 1) ], th [] [ text name ] ]
  ++ (List.map (\b -> td [class (if b then "cell-good" else "cell-bad")] [text (if b then "✓" else "✗")])
    <| List.map (\field -> not <| List.member field fields) watchedFields)

